package com.bergerkiller.bukkit.nolagg.meta;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import net.minecraft.server.ChunkProviderServer;
import net.minecraft.server.WorldServer;

import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.craftbukkit.entity.CraftEntity;
import org.bukkit.entity.Entity;
import org.bukkit.entity.ExperienceOrb;
import org.bukkit.entity.Item;

import com.all.bukkit.threadcheck.ThreadCheck;
import com.bergerkiller.bukkit.common.utils.CommonUtil;
import com.bergerkiller.bukkit.common.utils.MathUtil;
import com.bergerkiller.bukkit.nolagg.ItemHandler;
import com.bergerkiller.bukkit.nolagg.PerformanceMonitor;
import com.bergerkiller.bukkit.nolagg.WorldStackFormer;
import com.bergerkiller.bukkit.nolagg.spawnlimit.SpawnHandler;

public class WorldMetaData {

	private static final List<WorldMetaData> worlds = new ArrayList<WorldMetaData>();
	private static final HashMap<UUID, WorldMetaData> worlddata = new HashMap<UUID, WorldMetaData>();
	public static void deinit() {
		synchronized (worlddata) {
			for (WorldMetaData dat : worlddata.values()) {
				dat.remove();
			}
			worlddata.clear();
			worlds.clear();
		}
	}
	public static void remove(final World world) {
		synchronized (worlddata) {
			WorldMetaData dat = worlddata.remove(world.getUID());
			if (dat != null) {
				worlds.remove(dat);
				dat.remove();
			}
		}
	}
	public static WorldMetaData get(final World world) {
		if (world == null) return null;
		synchronized (worlddata) {
			WorldMetaData dat = worlddata.get(world.getUID());
			if (dat == null) {
				dat = new WorldMetaData(world);
				worlddata.put(world.getUID(), dat);
				worlds.add(dat);
			}
			return dat;
		}
	}
	public static WorldMetaData get(final net.minecraft.server.World world) {
		return get(world.getWorld());
	}
	public static List<WorldMetaData> getWorlds() {
		return worlds;
	}
	public static void updateAll() {
		if (!ThreadCheck.check(null)) return;
		SpawnHandler.reset();
		synchronized (worlddata) {
			PerformanceMonitor.entityListGen.start();
			for (WorldMetaData dat : worlds) dat.updateEntities();	
			PerformanceMonitor.entityListGen.next();
					
			PerformanceMonitor.entitySpawnLimit.start();	
			for (WorldMetaData dat : worlds) dat.updateSpawnLimits();
			PerformanceMonitor.entitySpawnLimit.next();
					
			PerformanceMonitor.itemStack.start();
			for (WorldMetaData dat : worlds) dat.updateStackForming();	
			PerformanceMonitor.itemStack.next();

			PerformanceMonitor.chunkUpdate.start();
			for (WorldMetaData dat : worlds) dat.updateChunks();
			PerformanceMonitor.chunkUpdate.next();
		}
	}
	private void updateEntities() {
		//update entity list
		this.entitylist.clear();
		this.itemlist.clear();
		this.orblist.clear();
		CraftEntity e;
		for (Object obj : this.getNative().entityList) {
			e = (CraftEntity) ((net.minecraft.server.Entity) obj).getBukkitEntity();
			if (e.isDead()) continue;
			if (e instanceof Item) {
				if (ItemHandler.isIgnoredItem(e)) continue;
				if (this.itemlist != null) this.itemlist.add((Item) e);
			} else if (e instanceof ExperienceOrb) {
				if (this.orblist != null) this.orblist.add((ExperienceOrb) e);
			}
			if (this.entitylist != null) this.entitylist.add(e);
		}
		Collections.reverse(this.entitylist); //make sure the oldest entity is at front
	}
	private void updateSpawnLimits() {
		SpawnHandler.update(this.world, this.entitylist);
	}
	private void updateStackForming() {
		this.stackFormer.update(this.itemlist, this.orblist);
	}
	public void updateStackFormingAsync() {
		this.stackFormer.run();
	}
	
	public static int getSpawnChunkX(WorldServer world) {
		return world.worldData.c() >> 4;
	}
	public static int getSpawnChunkZ(WorldServer world) {
		return world.worldData.e() >> 4;
	}
	public int getSpawnChunkX() {
		return getSpawnChunkX(this.getNative());
	}
	public int getSpawnChunkZ() {
		return getSpawnChunkZ(this.getNative());
	}
	
	private void updateChunks() {
		//update chunks and unload them if needed
		synchronized (this.chunks) {
			if (!this.chunks.isEmpty()) {
				final long time = System.currentTimeMillis() - ChunkMetaData.unloadDelay;
				Iterator<ChunkMetaData> it = this.chunks.iterator();
				ChunkMetaData dat;
				WorldServer world = this.getNative();
				while (it.hasNext()) {
					dat = it.next();
					if (!dat.isLoaded()) {
						it.remove();
						this.chunkdata.remove(dat.toLongCoordinate());
					} else if (world.savingDisabled || !dat.isUnloadable() || dat.isPlayerNear()) {
						dat.update();
						dat.touch();
					} else if (dat.isExpired(time)) {
						this.getChunkProvider().queueUnload(dat.getX(), dat.getZ());
					} else {
						dat.update();
					}
				}

			}
		}
	}

	public static int getTotalChunkCount() {
		int c = 0;
		synchronized (worlddata) {
			for (WorldMetaData wmd : worlddata.values()) c += wmd.getChunkCount();	
		}
		return c;
	}
	public static int getTotalUnloadableChunkCount() {
		int c = 0;
		synchronized (worlddata) {
			for (WorldMetaData wmd : worlddata.values()) {
				synchronized (wmd.chunkdata) {
					for (ChunkMetaData cmd : wmd.chunks) {
						if (cmd.isUnloadable()) c++;
					}
				}
			}
		}
		return c;
	}
	public static int getTotalHiddenItemCount() {
		int c = 0;
		synchronized (worlddata) {
			for (WorldMetaData wmd : worlddata.values()) {
				synchronized (wmd.chunkdata) {
					for (ChunkMetaData cmd : wmd.chunks) {
						c += cmd.getHiddenItemCount();
					}
				}
			}
		}
		return c;
	}

	public static void clearAllItems() {
		synchronized (worlddata) {
			for (WorldMetaData dat : worlddata.values()) dat.clearItems();
		}
	}
	public static boolean handleItemSpawn(Item item, boolean doIgnoreCheck) {
		ChunkMetaData dat = ChunkMetaData.get(item.getLocation());
		if (dat != null) return dat.handleItemSpawn(item, doIgnoreCheck);
		return true;
	}
	public static boolean removeSpawnedItem(Item item) {
		if (ItemHandler.maxItemsPerChunk <= 0) return false;
		WorldMetaData dat = get(item.getWorld());
		ChunkMetaData chunk = dat.getChunk(item.getLocation());
		if (chunk == null) return false;
		if (chunk.removeSpawnedItem(item)) {
			return true;
		} else {
			//loop through chunks
			synchronized (dat.chunkdata) {
				for (ChunkMetaData cdat : dat.chunks) {
					if (cdat.removeSpawnedItem(item)) return true;
				}
			}
			return false;
		}
	}

	private WorldMetaData(final World world) {
		this.world = (CraftWorld) world;
		this.chunkdata = new HashMap<Long, ChunkMetaData>(CommonUtil.chunkArea);
		this.chunks = new ArrayList<ChunkMetaData>(CommonUtil.chunkArea);
		this.entitylist = new ArrayList<Entity>();
		this.itemlist = new ArrayList<Item>();
		this.orblist = new ArrayList<ExperienceOrb>();
		this.add();
	}

	public final CraftWorld world;
	private final Map<Long, ChunkMetaData> chunkdata;
	private final List<ChunkMetaData> chunks;
	private final List<Entity> entitylist;
	private final List<Item> itemlist;
	private final List<ExperienceOrb> orblist;
	private final WorldStackFormer stackFormer = new WorldStackFormer();

	public ChunkProviderServer getChunkProvider() {
		return this.getNative().chunkProviderServer;
	}
	public WorldServer getNative() {
		return this.world.getHandle();
	}

	public int getHiddenItemCount() {
		int c = 0;
		synchronized (this.chunkdata) {
			for (ChunkMetaData cmd : this.chunks) {
				c += cmd.getHiddenItemCount();
			}
		}
		return c;
	}
	public void clearItems() {
		if (!ThreadCheck.check(null)) return;
		synchronized (this.chunkdata) {
			for (ChunkMetaData cmd : this.chunks) {
				cmd.clearItems();
			}
		}
	}

	public List<Entity> getEntities() {
		return this.entitylist;
	}
	public List<Item> getItems() {
		return this.itemlist;
	}
	public List<ExperienceOrb> getOrbs() {
		return this.orblist;
	}

	public ChunkMetaData getChunk(final Location location) {
		return this.getChunk(location.getBlockX() >> 4, location.getBlockZ() >> 4);
	}
	public ChunkMetaData getChunk(final net.minecraft.server.Chunk chunk) {
		return this.getChunk(chunk.x, chunk.z);
	}
	public ChunkMetaData getChunk(final Chunk chunk) {
		return this.getChunk(chunk.getX(), chunk.getZ());
	}
	public ChunkMetaData getChunk(final int x, final int z) {
		final long key = MathUtil.toLong(x, z);
		synchronized (this.chunkdata) {
			ChunkMetaData dat = this.chunkdata.get(key);
			if (dat == null && this.getChunkProvider().isChunkLoaded(x, z)) {
				dat = new ChunkMetaData(this, x, z);
				this.chunkdata.put(key, dat);
				this.chunks.add(dat);
			}
			return dat;
		}
	}
	public int getChunkCount() {
		synchronized (this.chunkdata) {
			return this.chunks.size();
		}
	}

	protected void removeChunk(final ChunkMetaData dat) {
		synchronized (this.chunkdata) {
			this.chunkdata.remove(dat.toLongCoordinate());
			this.chunks.remove(dat);
		}
	}
	public boolean removeChunk(final int x, final int z) {
		synchronized (this.chunkdata) {
			ChunkMetaData dat = this.chunkdata.get(MathUtil.toLong(x, z));
			if (dat != null) {
				dat.remove();
				return true;
			} else {
				return false;
			}
		}
	}
		
	private void add() {
		for (Chunk c : world.getLoadedChunks()) getChunk(c);
	}
	private void remove() {
		synchronized (this.chunkdata) {
			Iterator<ChunkMetaData> it = this.chunkdata.values().iterator();
			ChunkMetaData dat;
			while (it.hasNext()) {
				dat = it.next();
				dat.remove(false);
				it.remove();
			}
		}
		//other unload code here
	}

}
